#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io


from .maxwell2d import *


class Maxwell2DPeriodic(Maxwell2D):
    def __init__(self, *args, propagation_constant=0.0, **kwargs):
        super().__init__(*args, **kwargs)
        self.propagation_constant = propagation_constant

        if self.modal:
            self.propagation_vector = np.array([self.propagation_constant, 0])

        else:
            self.propagation_vector = self.source.wavenumber * np.array(
                [-np.sin(self.source.angle), -np.cos(self.source.angle)]
            )
        self.phasor = phasor(
            self.propagation_vector[0],
            direction=0,
            degree=self.degree,
            domain=self.geometry.mesh,
        )
        self.annex_field = (
            make_stack(
                self.geometry,
                self.coefficients,
                self.source,
                polarization=self.polarization,
                source_domains=self.source_domains,
                degree=self.degree,
                dim=2,
            )
            if not self.modal
            else None
        )

    @property
    def weak(self):
        u1 = self.annex_field["as_subdomain"]["stack"] if not self.modal else 0
        u = self.trial * self.phasor
        v = self.test * self.phasor.conj
        return super()._weak(u, v, u1)

    def build_boundary_conditions(self):

        applied_function = (
            Constant(0)
            if self.modal
            else -self.annex_field["as_subdomain"]["stack"] * self.phasor.conj
        )
        self._boundary_conditions = self.build_pec_boundary_conditions(applied_function)
        return self._boundary_conditions



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io

import glob
import os

import numpy as np
from scipy.constants import c, epsilon_0, mu_0

from .. import ADJOINT, dolfin
from ..bc import *
from ..complex import *
from ..formulations import *
from ..geometry import *
from ..materials import *
from ..sources import *
from ..utils import project_iterative
from ..utils.helpers import array2function


class Simulation:
    def __init__(self, geometry, formulation=None, direct=True):
        self.geometry = geometry
        self.formulation = formulation
        self.coefficients = formulation.coefficients
        self.function_space = formulation.function_space
        self._source = formulation.source
        self.boundary_conditions = formulation.boundary_conditions
        self.mesh = self.geometry.mesh
        self._boundary_conditions = []
        self.dx = formulation.dx
        self.ds = formulation.ds
        self.dS = formulation.dS
        self.direct = direct

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, value):
        self.formulation.source = value
        self._source = value

    def assemble_lhs(self):
        """Assemble the left hand side of the weak formulation.

        Returns
        -------
        PETSc matrix
            Assembled matrix.

        """
        self.formulation.build_lhs()
        self.matrix = assemble(self.formulation.lhs)
        return self.matrix

    def assemble_rhs(self, custom_rhs=None):
        """Assemble the right hand side of the weak formulation.

        Returns
        -------
        PETSc vector
            Assembled vector.

        """
        if custom_rhs is not None:
            self.formulation._set_rhs(custom_rhs)
        else:
            self.formulation.build_rhs()
        self.vector = assemble(self.formulation.rhs)
        return self.vector

    def assemble(self):
        """Assemble the weak formulation.

        Returns
        -------
        tuple (PETSc matrix, PETSc vector)
            The assembled matix and vector.

        """
        self.matrix = self.assemble_lhs()
        self.vector = self.assemble_rhs()
        return self.matrix, self.vector

    def apply_boundary_conditions(self):
        """Apply boundary conditions."""
        bcs = self.formulation.build_boundary_conditions()
        for bc in bcs:
            bc.apply(self.matrix, self.vector)

    def solve_system(self, again=False, vector_function=True):
        """Solve the discretized system.

        Parameters
        ----------
        again : bool
            Reuse solver (the default is False).
        vector_function : bool
            Use a vector function (the default is True).

        Returns
        -------
        dolfin Function
            The solution.

        """
        if vector_function:
            element = self.function_space.split()[0].ufl_element()
            V_vect = dolfin.VectorFunctionSpace(
                self.mesh, element.family(), element.degree()
            )
            u = dolfin.Function(V_vect)
        else:
            u = dolfin.Function(self.function_space)

        if not again:
            if self.direct:
                self.solver = dolfin.LUSolver(self.matrix, "mumps")
            else:
                # self.solver = dolfin.KrylovSolver(self.matrix,"cg", "jacobi")
                self.solver = dolfin.KrylovSolver(self.matrix)
        self.solver.solve(u.vector(), self.vector)
        solution = Complex(*u.split())
        return solution

    def solve(self):
        """Assemble, apply boundary conditions and computes the solution.

        Returns
        -------
        dolfin Function
            The solution.

        """

        self.assemble()
        self.apply_boundary_conditions()
        return self.solve_system()

    def eigensolve(self, n_eig=6, wavevector_target=0.0, tol=1e-6, **kwargs):

        wf = self.formulation.weak
        if self.formulation.dim == 1:
            dummy_vector = (
                dolfin.Constant(0) * self.formulation.test * self.formulation.dx
            )
        else:
            dummy_vector = (
                dot(dolfin.Constant((0, 0, 0)), self.formulation.test)
                * self.formulation.dx
            )

        dv = dummy_vector.real + dummy_vector.imag

        # Assemble matrices
        A = dolfin.PETScMatrix()
        B = dolfin.PETScMatrix()
        b = dolfin.PETScVector()

        bcs = self.formulation.build_boundary_conditions()

        # [bc.zero(B) for bc in bcs]

        dolfin.assemble_system(wf[0], dv, bcs, A_tensor=A, b_tensor=b)
        dolfin.assemble_system(wf[1], dv, bcs, A_tensor=B, b_tensor=b)

        eigensolver = dolfin.SLEPcEigenSolver(
            dolfin.as_backend_type(A), dolfin.as_backend_type(B)
        )
        # eigensolver.parameters["problem_type"] = "gen_hermitian"
        # eigensolver.parameters["spectrum"] = "target real"
        eigensolver.parameters["spectrum"] = "target magnitude"
        eigensolver.parameters["solver"] = "krylov-schur"
        # eigensolver.parameters["solver"] = "power"
        eigensolver.parameters["spectral_shift"] = float(wavevector_target ** 2)
        eigensolver.parameters["spectral_transform"] = "shift-and-invert"
        eigensolver.parameters["tolerance"] = tol
        # eigensolver.parameters["solver"] = "mumps"
        dolfin.PETScOptions.set("st_ksp_type", "preonly")
        dolfin.PETScOptions.set("st_pc_type", "lu")
        # dolfin.PETScOptions.set("st_pc_factor_mat_solver_type", "mumps")
        # dolfin.PETScOptions.set("eps_max_it", "300")
        # dolfin.PETScOptions.set("eps_target", "0.00001")
        # dolfin.PETScOptions.set("eps_mpd", "600")
        # dolfin.PETScOptions.set("eps_nev", "400")
        eigensolver.parameters.update(kwargs)
        eigensolver.solve(2 * n_eig)

        nconv = eigensolver.get_number_converged()

        self.solution = {}
        self.solution["converged"] = nconv

        KNs = []
        UNs = []

        nconv = min(2 * n_eig, nconv)

        for j in range(nconv):
            ev_re, ev_im, rx, cx = eigensolver.get_eigenpair(j)
            eig_vec_re = array2function(rx, self.formulation.function_space)
            eig_vec_im = array2function(cx, self.formulation.function_space)

            if self.formulation.dim == 1:
                eig_vec = Complex(*eig_vec_re) + 1j * Complex(*eig_vec_im)
            else:

                eig_vec_re_re = [eig_vec_re[i] for i in range(3)]
                eig_vec_re_im = [eig_vec_re[i] for i in range(3, 6)]
                eig_vec_im_re = [eig_vec_im[i] for i in range(3)]
                eig_vec_im_im = [eig_vec_im[i] for i in range(3, 6)]

                re = Complex(eig_vec_re_re, eig_vec_re_im)
                im = Complex(eig_vec_im_re, eig_vec_im_im)
                eig_vec = vector(re) + 1j * vector(im)

            # eig_vec = Complex(eig_vec_re[0],eig_vec_im[1])
            ev = ev_re + 1j * ev_im
            kn = (ev) ** 0.5
            KNs.append(kn)
            UNs.append(eig_vec)

        # HACK: We get the complex conjugates as well so compute twice
        # as much eigenvalues and return only half
        KNs = np.array(KNs)[::2]
        UNs = UNs[::2]

        self.solution["eigenvalues"] = KNs
        self.solution["eigenvectors"] = UNs
        self.eigensolver = eigensolver
        return self.solution



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io


from .metaclasses import _ScatteringBase
from .simulation import *


class Scatt2D(_ScatteringBase, Simulation):
    def __init__(
        self,
        geometry,
        epsilon,
        mu,
        source=None,
        boundary_conditions={},
        polarization="TM",
        modal=False,
        degree=1,
        pml_stretch=1 - 1j,
        element="CG",
    ):
        assert isinstance(geometry, BoxPML2D)
        if source is not None:
            assert source.dim == 2
        self.element = element
        function_space = ComplexFunctionSpace(geometry.mesh, self.element, degree)
        pmlx = PML(
            "x", stretch=pml_stretch, matched_domain="box", applied_domain="pmlx"
        )
        pmly = PML(
            "y", stretch=pml_stretch, matched_domain="box", applied_domain="pmly"
        )
        pmlxy = PML(
            "xy", stretch=pml_stretch, matched_domain="box", applied_domain="pmlxy"
        )

        epsilon_coeff = Coefficient(
            epsilon, geometry, pmls=[pmlx, pmly, pmlxy], degree=degree
        )
        mu_coeff = Coefficient(mu, geometry, pmls=[pmlx, pmly, pmlxy], degree=degree)

        coefficients = epsilon_coeff, mu_coeff
        no_source_domains = ["box", "pmlx", "pmly", "pmlxy"]
        if modal:
            source_domains = []
        else:
            source_domains = [
                dom for dom in geometry.domains if dom not in no_source_domains
            ]
        formulation = Maxwell2D(
            geometry,
            coefficients,
            function_space,
            source=source,
            source_domains=source_domains,
            reference="box",
            polarization=polarization,
            modal=modal,
            boundary_conditions=boundary_conditions,
        )

        super().__init__(geometry, formulation)

        self.degree = degree

    def solve_system(self, again=False):
        u = super().solve_system(again=again, vector_function=False)
        self.solution = {}
        self.solution["diffracted"] = u
        self.solution["total"] = u + self.source.expression
        return u

    @property
    def time_average_incident_poynting_vector_norm(self):
        Z0 = np.sqrt(mu_0 / epsilon_0)
        S0 = 1 / (2 * Z0) if self.formulation.polarization == "TM" else 0.5 * Z0
        return S0

    def _cross_section_helper(self, return_type="s", boundaries="calc_bnds"):
        uscatt = self.solution["diffracted"]
        vscatt = self.formulation.get_dual(uscatt)
        utot = self.solution["total"]
        vtot = self.formulation.get_dual(utot)
        ui = self.source.expression
        vi = self.formulation.get_dual(ui)

        parallel = dolfin.MPI.comm_world.size > 1

        ### normal vector is messing up in parallel so workaround here:
        if parallel:
            n_out = dolfin.Expression(
                (f"x[0]/Rcalc", f"x[1]/Rcalc"),
                degree=self.degree,
                Rcalc=self.geometry.Rcalc,
            )
        else:
            n_out = self.geometry.unit_normal_vector

        dcalc = self.dS(boundaries)

        if return_type == "s":
            Sscatt = _get_power_dens(uscatt, vscatt)
            W = assemble(dot(n_out, Sscatt)("+") * dcalc)
        elif return_type == "a":
            Stot = _get_power_dens(utot, vtot)
            W = assemble(dot(n_out, Stot)("+") * dcalc)
        else:
            Se = _get_power_dens(uscatt, vi) + _get_power_dens(ui, vscatt)
            W = assemble(dot(n_out, Se)("+") * dcalc)
        out = W / self.time_average_incident_poynting_vector_norm
        return out if out > 0 else -out

    def scattering_cross_section(self, **kwargs):
        return self._cross_section_helper("s", **kwargs)

    def extinction_cross_section(self, **kwargs):
        return self._cross_section_helper("e", **kwargs)

    def absorption_cross_section(self, **kwargs):
        return self._cross_section_helper("a", **kwargs)

    def local_density_of_states(self, x, y):
        """Compute the local density of state.

        Parameters
        ----------
        x : float
            x coordinate.
        y : float
            y coordinate.

        Returns
        -------
        float
            The local density of states.

        """
        # self.source =  LineSource(
        #     wavelength=self.source.wavelength, position=(x, y),domain=self.mesh, degree=self.degree,
        # )
        self.source.position = x, y
        if hasattr(self, "solution"):
            self.assemble_rhs()
            self.solve_system(again=True)
        else:
            self.solve()
        u = self.solution["total"]
        eps = dolfin.DOLFIN_EPS_LARGE
        delta = 1 + eps
        evalpoint = x * delta, y * delta
        if evalpoint[0] == 0:
            evalpoint = eps, evalpoint[1]
        if evalpoint[1] == 0:
            evalpoint = evalpoint[0], eps
        ldos = -2 * self.source.pulsation / (np.pi * c ** 2) * u(evalpoint).imag
        return ldos

    def check_optical_theorem(cs, rtol=1e-12):
        np.allclose(cs["extinction"], cs["scattering"] + cs["absorption"], rtol=rtol)

    def plot_field(
        self,
        type="real",
        field="total",
        ax=None,
        mincmap=None,
        maxcmap=None,
        fig=None,
        phase=0,
        callback=None,
        **kwargs,
    ):

        u = self.solution[field]
        if ax == None:
            ax = plt.gca()
        if "cmap" not in kwargs:
            kwargs["cmap"] = "RdBu_r"

        f = u * phase_shift(phase, degree=self.degree)

        fplot = check_plot_type(type, f)
        fplot = project_iterative(
            fplot,
            self.formulation.real_function_space,
        )
        pp = dolfin.plot(fplot, **kwargs)

        ppmax = pp.cvalues[-1]
        ppmin = pp.cvalues[0]
        ax.set_aspect(1)
        bs = self.geometry.box_size
        ax.set_xlim(-bs[0] / 2, bs[0] / 2)
        ax.set_ylim(-bs[1] / 2, bs[1] / 2)
        mincmap = mincmap or ppmin
        maxcmap = maxcmap or ppmax
        pp.set_clim(mincmap, maxcmap)

        cm = plt.cm.ScalarMappable(cmap=kwargs["cmap"])
        cm.set_clim(mincmap, maxcmap)

        fig = plt.gcf() if fig is None else fig
        cb = fig.colorbar(cm, ax=ax)

        if callback is not None:
            callback(**kwargs)

        return pp, cb

    def animate_field(self, n=11, filename="animation.gif", **kwargs):
        import tempfile

        from PIL import Image

        from ..plot import plt

        anim = []
        tmpdir = tempfile.mkdtemp()
        fp_in = f"{tmpdir}/animation_tmp_*.png"
        phase = np.linspace(0, 2 * np.pi, n + 1)[:n]
        for iplot in range(n):
            number_str = str(iplot).zfill(4)
            pngname = f"{tmpdir}/animation_tmp_{number_str}.png"
            p = self.plot_field(phase=phase[iplot], **kwargs)
            fig = plt.gcf()
            fig.savefig(pngname)
            fig.clear()
            anim.append(p)

        plt.close(fig)

        img, *imgs = [Image.open(f) for f in sorted(glob.glob(fp_in))]
        img.save(
            fp=filename,
            format="GIF",
            append_images=imgs,
            save_all=True,
            duration=200,
            loop=0,
        )
        os.system(f"rm -f {tmpdir}/animation_tmp_*.png")
        return anim


def _get_power_dens(u, v):
    re = as_vector([v[1].real, -v[0].real])
    im = as_vector([v[1].imag, -v[0].imag])
    EcrossHstar = u * Complex(re, -im)
    return (Constant(0.5) * EcrossHstar).real



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io


"""Metaclasses"""


from abc import ABC, abstractmethod


class _ScatteringBase(ABC):
    """Scattering problem."""

    # def __init__(self,*args,**kwargs):
    #     pass

    @abstractmethod
    def scattering_cross_section(self):
        """Compute the scattering cross section.

        Returns
        -------
        float
            Scattering cross section.

        """
        pass

    @abstractmethod
    def absorption_cross_section(self):
        """Compute the absorption cross section.

        Returns
        -------
        float
            Absorption cross section.

        """
        pass

    @abstractmethod
    def extinction_cross_section(self):
        """Compute the extinction cross section.

        Returns
        -------
        float
            Extinction cross section.

        """
        pass

    def get_cross_sections(self, **kwargs):
        """Compute cross sections.

        Returns
        -------
        dict
            A dictionary containing scattering, absorption and extinction
            cross sections.

        """

        scs = self.scattering_cross_section(**kwargs)
        acs = self.absorption_cross_section(**kwargs)
        ecs = self.extinction_cross_section(**kwargs)
        return dict(scattering=scs, absorption=acs, extinction=ecs)


class _GratingBase(ABC):
    """Base class for grating problems."""

    @abstractmethod
    def diffraction_efficiencies(
        self,
        N_order=0,
        cplx_effs=False,
        orders=False,
        subdomain_absorption=False,
        verbose=False,
    ):
        """Compute the diffraction efficiencies.

        Parameters
        ----------
        N_order : int
            Number of diffraction orders (the default is 0). This will include negative
            and positive orders. In 2D, it will calculate `2 * N_order + 1` coefficients,
            and `(2 * N_order + 1) ** 2` in 3D.
        cplx_effs : bool
            If `True`, return complex coefficients (amplitude reflection and transmission).
            If `False`, return real coefficients (power reflection and transmission)
        orders : bool
            If `True`, computes the transmission and reflection per diffraction orders.
            If `False`, returns the sum of the diffraction orders.
            (the default is False).
        subdomain_absorption : bool
            If `True`, computes the absorption per subdomains and splits the electric and magnetic contributions.
            If `False`, returns the total absorption.
            (the default is False).
        verbose : bool
            If `True`, prints the energy balance (the default is False).

        Returns
        -------
        dict {`R`, `T`, `Q`, `B`}
            A dictionary containing reflection `R`, transmission, `T`, absorption, `Q`, and energy balance `B`.

        """

        pass

    @abstractmethod
    def compute_absorption(self, subdomain_absorption=False):
        """Computes the absorption.

        Parameters
        ----------
        subdomain_absorption : bool
            If `True`, computes the absorption per subdomain and splits the electric and magnetic contributions.
            If `False`, returns the total absorption.
            (the default is False).

        Returns
        -------
        Q_tot : float
            Total absorption.
        Q_domains : dict or dict of dict
            Detailled absorption with electric and magnetic contributions.
            Returns aborption per subdomain if `subdomain_absorption=True`.

        """
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io


__all__ = [
    "Layered",
    "Lattice",
    "Scattering",
    "Grating",
    "PhotonicCrystal",
    "PlaneWave",
    "LineSource",
    "GaussianBeam",
    "BoxPML",
    "Layered",
    "Geometry",
    "Homogenization2D",
]


from .geometry import *
from .models import *
from .models.metaclasses import _GratingBase, _ScatteringBase
from .sources import *


def _check_dimension(dim):
    if dim not in [2, 3]:
        raise ValueError("dimension must be 2 or 3")


class BoxPML(Geometry):
    """BoxPML(dim, box_size, box_center, pml_width, Rcalc=0, model_name="Box with PMLs", **kwargs)
    Computational domain with Perfectly Matched Layers (PMLs).

    Parameters
    ----------
    dim : int
        Geometric dimension (either 2 or 3, the default is 3).
    box_size : tuple of floats of length `dim`
        Size of the box: :math:`(l_x,l_y)` if `dim=2` or :math:`(l_x, l_y, l_z)` if `dim=3`.
    box_center : tuple of floats of length `dim`
        Size of the box: :math:`(c_x,c_y)` if `dim=2` or :math:`(c_x, c_y, c_z)` if `dim=3`.
    pml_width : tuple of floats of length `dim`
        Size of the PMLs: :math:`(h_x,h_y)` if `dim=2` or :math:`(h_x, h_y, h_z)` if `dim=3`.
    model_name : str
        Name of the model.
    **kwargs : dictionary
        Additional parameters. See the parent class :class:`~gyptis.Geometry`.

    """

    def __new__(self, dim=3, *args, **kwargs):
        if dim not in [2, 3]:
            raise ValueError("dimension must be 2 or 3")
        if dim == 3:
            return BoxPML3D(*args, **kwargs)
        else:
            return BoxPML2D(*args, **kwargs)


class Layered(Geometry):
    """Layered(dim, period, thicknesses, **kwargs)
    Layered media for diffraction problems, defining the periodic unit cell
    for mono or bi periodic gratings.

    Parameters
    ----------
    dim : int
        Geometric dimension (either 2 or 3, the default is 3).
    period : float or tuple
        In 2D, periodicity of the grating :math:`d` along :math:`x` (float).
        In 3D, periodicity of the grating :math:`(d_x,d_y)` along :math:`x`
        and :math:`y` (tuple of floats of lenght 2).
    thicknesses : :class:`~collections.OrderedDict`
        Dictionary containing physical names and thicknesses from top to bottom.
        (``thicknesses["phyiscal_name"]=thickness_value``)
    **kwargs : dictionary
        Additional parameters. See the parent class :class:`~gyptis.Geometry`.


    Examples
    --------

    >>> from collections import OrderedDict
    >>> from gyptis import Layered
    >>> t = OrderedDict(pml_top=1, slab=3, pml_bot=1)
    >>> lays = Layered(dim=2, period=1.3, thicknesses=t)
    >>> lays.build()

    """

    def __new__(self, dim=3, *args, **kwargs):
        _check_dimension(dim)
        if dim == 3:
            return Layered3D(*args, **kwargs)
        else:
            return Layered2D(*args, **kwargs)


class Lattice(Geometry):
    """Lattice(vectors, **kwargs)
    Unit cell for periodic problems.

    Parameters
    ----------
    dim : int
        Geometric dimension (either 2 or 3, the default is 3).
    vectors : tuple
        In 2D, a tuple of lengh 2 with the :math:`(x,y)` coordinates of 2 basis vectors.
        In 3D, a tuple of lengh 3 with the :math:`(x,y,z)` coordinates of 3 basis vectors.
    **kwargs : dictionary
        Additional parameters. See the parent class :class:`~gyptis.Geometry`.
    """

    def __new__(self, dim=3, *args, **kwargs):
        _check_dimension(dim)
        if dim == 3:
            raise NotImplementedError
        else:
            return Lattice2D(*args, **kwargs)


class Scattering(_ScatteringBase, Simulation):
    """Scattering(geometry, epsilon, mu, source=None, boundary_conditions={}, polarization="TM", modal=False, degree=1, pml_stretch=1 - 1j)
    Scattering problem.

    Parameters
    ----------
    geometry : :class:`~gyptis.Geometry`
        The meshed geometry
    epsilon : dict
        Permittivity in various subdomains.
    mu : dict
        Permeability in various subdomains.
    source : :class:`~gyptis.source.Source`
        Excitation (the default is None).
    boundary_conditions : dict
        Boundary conditions {"boundary": "condition"} (the default is {}).
        Valid condition is only "PEC".
    polarization : str
        Polarization case (only makes sense for 2D problems, the default is "TM").
    modal : str
        Perform modal analysis (the default is False).
    degree : int
        Degree of finite elements interpolation (the default is 1).
    pml_stretch : complex
        Complex coordinate stretch for te PMLs (the default is 1 - 1j).

    """

    def __new__(self, *args, **kwargs):
        geom = kwargs.get("geometry") or args[0]
        _check_dimension(geom.dim)
        if geom.dim == 3:
            return Scatt3D(*args, **kwargs)
        else:
            return Scatt2D(*args, **kwargs)


class Grating(_GratingBase, Simulation):
    """Grating(geometry, epsilon, mu, source, boundary_conditions={}, polarization="TM", degree=1, pml_stretch=1 - 1j, periodic_map_tol=1e-8, propagation_constant=0.0)
    Grating problem.

    Parameters
    ----------
    geometry : :class:`~gyptis.Geometry`
        The meshed geometry
    epsilon : dict
        Permittivity in various subdomains.
    mu : dict
        Permeability in various subdomains.
    source : :class:`~gyptis.source.Source`
        Excitation (the default is None).
    boundary_conditions : dict
        Boundary conditions {"boundary": "condition"} (the default is {}).
        Valid condition is only "PEC".
    polarization : str
        Polarization case (only makes sense for 2D problems, the default is "TM").
    modal : str
        Perform modal analysis (the default is False).
    degree : int
        Degree of finite elements interpolation (the default is 1).
    pml_stretch : complex
        Complex coordinate stretch for te PMLs (the default is 1 - 1j).
    periodic_map_tol : float
        Tolerance for mapping boundaries (the default is 1e-8).
    propagation_constant : float
        Propagation constant along the periodicity. Only
        makes sense for modal analysis (the default is 0.0).


    """

    def __new__(self, *args, **kwargs):
        geom = kwargs.get("geometry") or args[0]
        _check_dimension(geom.dim)
        if geom.dim == 3:
            return Grating3D(*args, **kwargs)
        else:
            return Grating2D(*args, **kwargs)


class PhotonicCrystal:
    def __new__(self, *args, **kwargs):
        geom = kwargs.get("geometry") or args[0]
        _check_dimension(geom.dim)
        if geom.dim == 3:
            raise NotImplementedError
        else:
            return PhotonicCrystal2D(*args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io


from .formulation import *


class TwoScale2D(Formulation):
    def __init__(
        self,
        geometry,
        coefficients,
        function_space,
        boundary_conditions={},
        degree=1,
        direction="x",
        case="epsilon",
    ):
        super().__init__(
            geometry,
            coefficients,
            function_space,
            boundary_conditions=boundary_conditions,
            degree=degree,
        )

        self.epsilon, self.mu = self.coefficients
        self.direction = direction
        self.case = case

        if self.case == "mu":
            self.xi = self.mu.to_xi()
            self.chi = self.epsilon.to_chi()
        else:
            self.xi = self.epsilon.to_xi()
            self.chi = self.mu.to_chi()

        self.pec_boundaries = prepare_boundary_conditions(boundary_conditions)

    def poisson(self, u, v, xi, domain="everywhere"):
        e = Constant((1, 0)) if self.direction == "x" else Constant((0, 1))

        form = []
        form.append(inner(xi * grad(u), grad(v)))
        form.append(dot(xi * e, grad(v)))
        return (form[0] + form[1]) * self.dx(domain)

    def _weak(self, u, v):
        xi = self.xi.as_subdomain()

        xi_dict = self.xi.as_property()

        dom_func, dom_no_func = find_domains_function((self.xi, self.chi))

        form = self.poisson(u, v, xi, domain=dom_no_func)
        for dom in dom_func:
            form += self.poisson(u, v, xi_dict[dom], domain=dom)

        weak = form.real + form.imag
        return weak

    @property
    def weak(self):
        u = self.trial
        v = self.test
        return self._weak(u, v)

    def build_pec_boundary_conditions(self, applied_function):
        if self.case == "epsilon" and self.pec_boundaries != []:
            ## FIXME: project is slow, avoid it.
            applied_function = project_iterative(
                applied_function, self.real_function_space
            )
            _boundary_conditions = build_pec_boundary_conditions(
                self.pec_boundaries,
                self.geometry,
                self.function_space,
                applied_function,
            )
        else:
            _boundary_conditions = []

        return _boundary_conditions

    def build_boundary_conditions(self):
        applied_function = Constant(0)
        self._boundary_conditions = self.build_pec_boundary_conditions(applied_function)
        return self._boundary_conditions



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io


"""
Scattering Matrix calculations for a scatterer.

See:
Demésy, G., Auger, J.-C. & Stout, B. Scattering matrix of arbitrarily shaped
objects: combining finite elements and vector partial waves.
J. Opt. Soc. Am. A, JOSAA 35, 1401–1409 (2018)
"""

import scipy.special as sp

from .. import dolfin as df
from .. import pi
from ..complex import Complex, Constant, j, vector
from .source import Source


def sqrt_safe(x):
    x = Complex(x, 0)
    t = x.phase
    return df.sqrt(x.module) * Complex(df.cos(t / 2), df.sin(t / 2))


def sph_bessel_J(nu, f):
    return df.sqrt(pi / (2 * f)) * df.bessel_J(nu + 0.5, f)


def sph_bessel_Y(nu, f):
    return df.sqrt(pi / (2 * f)) * df.bessel_Y(nu + 0.5, f)


def sph_bessel_H1(nu, f):
    return Complex(sph_bessel_J(nu, f), sph_bessel_Y(nu, f))


def sph_bessel_H2(nu, f):
    return Complex(sph_bessel_J(nu, f), -sph_bessel_Y(nu, f))


def ricatti_bessel_J(nu, f):
    return f * sph_bessel_J(nu, f)


def ricatti_bessel_Y(nu, f):
    return f * sph_bessel_Y(nu, f)


def ricatti_bessel_H1(nu, f):
    return Complex(ricatti_bessel_J(nu, f), ricatti_bessel_Y(nu, f))


def ricatti_bessel_H2(nu, f):
    return Complex(ricatti_bessel_J(nu, f), -ricatti_bessel_Y(nu, f))


def sph_jn(nu, f):
    return (pi / (2 * f)) ** 0.5 * sp.jv(nu + 0.5, f)


def sph_yn(nu, f):
    return (pi / (2 * f)) ** 0.5 * sp.yv(nu + 0.5, f)


def sph_hn1(nu, f):
    return sph_jn(nu, f) + 1j * sph_yn(nu, f)


def sph_hn2(nu, f):
    return sph_jn(nu, f) - 1j * sph_yn(nu, f)


def rb_jn(nu, f):
    return f * sph_jn(nu, f)


def rb_yn(nu, f):
    return f * sph_yn(nu, f)


def rb_hn1(nu, f):
    return rb_jn(nu, f) + 1j * rb_yn(nu, f)


def rb_hn2(nu, f):
    return rb_jn(nu, f) - 1j * rb_yn(nu, f)


def nm2p(n, m):
    return n * (n + 1) - m


def p2nm(p):
    n = int(p ** 0.5)
    m = n * (n + 1) - p
    return n, m


# Recurrence relations for scalar spherical harmonics


def P(x, n, m, y=None):
    y = y or sqrt_safe(1 - x ** 2)
    if m < 0:
        return (-1) ** (-m) * P(x, n, -m)
    elif n == m == 0:
        return 1 / (4 * pi) ** 0.5
    elif n == m:
        return -(((2 * n + 1) / (2 * n)) ** 0.5) * y * P(x, n - 1, m - 1)
    elif m == (n - 1):
        return x * (2 * n + 1) ** 0.5 * P(x, n - 1, n - 1)
    else:
        return (((2 * n + 1) / (n ** 2 - m ** 2)) ** 0.5) * (
            (2 * n - 1) ** 0.5 * x * P(x, n - 1, m)
            - (((n - 1) ** 2 - m ** 2) / (2 * n - 3)) ** 0.5 * P(x, n - 2, m)
        )


def u(x, n, m, y=None):
    y = y or sqrt_safe(1 - x ** 2)
    if m < 0:
        return (-1) ** (-m + 1) * u(x, n, -m)
    elif m == 0:
        return 0
    elif m == 1:
        return -1 / 4 * (3 / pi) ** 0.5
    elif n == m:
        return (
            -((n * (2 * n + 1) / (2 * (n + 1) * (n - 1))) ** 0.5)
            * y
            * u(x, n - 1, n - 1)
        )
    elif m == (n - 1):
        return ((2 * n + 1) * (n - 1) / (n + 1)) ** 0.5 * x * u(x, n - 1, n - 1)
    else:
        return ((2 * n + 1) * (n - 1) / ((n + 1) * (n ** 2 - m ** 2))) ** 0.5 * (
            x * (2 * n - 1) ** 0.5 * u(x, n - 1, m)
            - ((n - 2) * ((n - 1) ** 2 - m ** 2) / (n * (2 * n - 3))) ** 0.5
            * u(x, n - 2, m)
        )


def s(x, n, m, y=None):
    y = y or sqrt_safe(1 - x ** 2)
    if m < 0:
        return (-1) ** (-m) * s(x, n, -m)
    elif n == m:
        return x * u(x, n, m)
    else:
        return 1 / (m + 1) * ((n + m + 1) * (n - m)) ** 0.5 * y * u(
            x, n, m + 1
        ) + x * u(x, n, m)


class SphericalSource(Source):
    def __init__(self, wavelength, tol=1e-14, **kwargs):
        super().__init__(wavelength, dim=3, **kwargs)
        self.tol = tol
        r3D = "sqrt(pow(x[0],2) + pow(x[1],2) + pow(x[2],2))"
        r2D = "sqrt(pow(x[0],2) + pow(x[1],2))"
        self.r = df.Expression(
            f"{r3D}",
            **kwargs,
        )
        self.cos_theta = df.Expression(
            f"x[2]/({self.tol} + {r3D})",
            **kwargs,
        )
        self.sin_theta = df.Expression(
            f"{r2D}/({self.tol} + {r3D})",
            **kwargs,
        )
        self.cos_phi = df.Expression(
            f"x[0]/({self.tol} + {r2D})",
            **kwargs,
        )
        self.sin_phi = df.Expression(
            f"x[1]/({self.tol} + {r2D})",
            **kwargs,
        )
        ct = self.cos_theta
        st = self.sin_theta
        cp = self.cos_phi
        sp = self.sin_phi
        r = self.r
        rot_mat = [
            [st * cp, ct * cp, -sp],
            [st * sp, ct * sp, cp],
            [ct, -st, df.Constant(0)],
        ]
        self.transfo_matrix = Complex(df.as_tensor(rot_mat))

    def phasor(self, m):
        if m == 0:
            return Constant(1 + 0j)
        else:
            return Complex(self.cos_phi, self.sin_phi) ** df.Constant(m)


class SphericalHarmonic(SphericalSource):
    def __init__(self, wavelength, n, m, **kwargs):
        super().__init__(wavelength, **kwargs)
        self.n = n
        self.m = m
        self.phm = self.phasor(self.m)
        self.components = {}


class SphericalX(SphericalHarmonic):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.components["r"] = Constant(0)
        self.components["theta"] = j * u(self.cos_theta, self.n, self.m) * self.phm
        self.components["phi"] = -s(self.cos_theta, self.n, self.m) * self.phm
        self.expression_spherical = vector(self.components.values())
        self.expression = self.transfo_matrix * self.expression_spherical


class SphericalY(SphericalHarmonic):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.components["r"] = P(self.cos_theta, self.n, self.m) * self.phm
        self.components["theta"] = Constant(0)
        self.components["phi"] = Constant(0)
        self.expression_spherical = vector(self.components.values())
        self.expression = self.transfo_matrix * self.expression_spherical


class SphericalZ(SphericalHarmonic):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.components["r"] = Constant(0)
        self.components["theta"] = s(self.cos_theta, self.n, self.m) * self.phm
        self.components["phi"] = j * u(self.cos_theta, self.n, self.m) * self.phm
        self.expression_spherical = vector(self.components.values())
        self.expression = self.transfo_matrix * self.expression_spherical


class SphericalM(SphericalHarmonic):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        k = df.Constant(self.wavenumber)
        jn = sph_bessel_J(self.n, k * self.r)
        Xnm = SphericalX(*args, **kwargs)
        self.expression_spherical = jn * Xnm.expression_spherical
        self.expression = self.transfo_matrix * self.expression_spherical


class SphericalN(SphericalHarmonic):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        k = df.Constant(self.wavenumber)
        jn = sph_bessel_J(self.n, k * self.r)
        psin_prime = ricatti_bessel_J(self.n - 1, k * self.r) - self.n * jn
        Y = SphericalY(*args, **kwargs)
        Z = SphericalZ(*args, **kwargs)
        Ynm = Y.expression_spherical
        Znm = Z.expression_spherical
        self.expression_spherical = (
            (self.n * (self.n + 1)) ** 0.5 * jn * Ynm + psin_prime * Znm
        ) / (k * self.r)
        self.expression = self.transfo_matrix * self.expression_spherical



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io

"""
Sources.
"""

import numpy as np
import sympy as sp
from dolfin import Constant as ConstantRe
from scipy.constants import c, epsilon_0, mu_0
from sympy.vector import CoordSys3D

from ..complex import Complex, Constant, as_tensor, as_vector, dolfin, dot, grad

_COORD = CoordSys3D("N")


def vector(components):
    return (
        components[0] * _COORD.i + components[1] * _COORD.j + components[2] * _COORD.k
    )


x = sp.symbols("x[0] x[1] x[2]", real=True)
X = vector(np.array(x))


def expression2complex_2d(expr, **kwargs):
    re, im = (p.subs(x[2], 0) for p in expr.as_real_imag())
    dexpr = [dolfin.Expression(sp.printing.ccode(p), **kwargs) for p in (re, im)]
    return Complex(*dexpr)


def green_function_2d(wavelength, xs, ys, amplitude=1, degree=1, domain=None):
    Xs = vector(sp.symbols("xs, ys, 0", real=True))
    k0 = sp.symbols("k0", real=True)
    Xshift = X - Xs
    rho = sp.sqrt(Xshift.dot(Xshift))
    rho = rho.subs(x[2], 0)
    kr = k0 * rho
    k0_ = 2 * np.pi / wavelength
    KR = dolfin.Expression(
        sp.printing.ccode(kr), k0=k0_, xs=xs, ys=ys, degree=degree, domain=domain
    )
    return (
        -1
        / 4
        * Complex(dolfin.bessel_Y(0, KR), dolfin.bessel_J(0, KR))
        * Constant(amplitude)
    )


class Source:
    def __init__(self, wavelength, dim, degree=1, domain=None):
        self.wavelength = wavelength
        self.dim = dim
        self.degree = degree
        self.domain = domain

    @property
    def wavenumber(self):
        return 2 * np.pi / self.wavelength

    @property
    def pulsation(self):
        return self.wavenumber * c

    @property
    def frequency(self):
        return c / self.wavelength



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io

__all__ = ["logger", "set_log_level"]


import logging
import sys

from loguru import logger

from .. import dolfin


def set_log_level(level):
    global logger
    logger.remove()
    format = "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> -- <level>{message}</level>"
    logger.add(sys.stderr, format=format, level=level, colorize=True)


set_log_level(logging.INFO)
dolfin.set_log_level(logging.INFO)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io


from .dipole import Dipole
from .gaussian import GaussianBeam
from .ls import LineSource
from .pw import *
from .source import *
from .spharm import *
from .stack import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io

from .source import *


class LineSource(Source):
    def __init__(self, wavelength, position, amplitude=1, degree=1, domain=None):
        super().__init__(wavelength, dim=2, degree=degree, domain=domain)
        self.position = position
        self.amplitude = amplitude

    @property
    def expression(self):
        _expression = green_function_2d(
            self.wavelength,
            *self.position,
            amplitude=self.amplitude,
            degree=self.degree,
            domain=self.domain,
        )
        return _expression



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io

from .grating2d import *
from .grating3d import *
from .phc2d import *
from .phc3d import *
from .scattering2d import *
from .scattering3d import *
from .simulation import *
from .twoscale2d import *
from .twoscale3d import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io


from .source import *


class GaussianBeam(Source):
    def __init__(
        self, wavelength, angle, waist, dim, Npw=101, amplitude=1, degree=1, domain=None
    ):
        super().__init__(wavelength, dim, degree=degree, domain=domain)
        self.angle = angle
        self.amplitude = amplitude
        self.Npw = Npw
        self.waist = waist

    @property
    def expression(self):
        if self.dim == 2:
            _expression = Constant(0)
            for t in np.linspace(-np.pi / 2, np.pi / 2, self.Npw):
                _expression += (
                    plane_wave_2d(
                        self.wavelength,
                        self.angle + t,
                        amplitude=self.amplitude,
                        degree=self.degree,
                        domain=self.domain,
                    )
                    * Constant(
                        np.exp(-(t ** 2) * 4 * (np.pi / 2) ** 2 * self.waist ** 2)
                    )
                )
            dk = np.pi / (self.Npw - 1)
            _expression *= Constant(dk)
        else:
            raise (NotImplementedError)
        return _expression



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io

from .simulation import *


class PhotonicCrystal2D(Simulation):
    def __init__(
        self,
        geometry,
        epsilon,
        mu,
        propagation_vector=(0, 0),
        boundary_conditions={},
        polarization="TM",
        degree=1,
    ):
        assert isinstance(geometry, Lattice2D)

        self.periodic_bcs = BiPeriodic2D(geometry)
        function_space = ComplexFunctionSpace(
            geometry.mesh, "CG", degree, constrained_domain=self.periodic_bcs
        )
        epsilon = {k: e + 1e-16j for k, e in epsilon.items()}
        mu = {k: m + 1e-16j for k, m in mu.items()}
        epsilon_coeff = Coefficient(epsilon, geometry, degree=degree)
        mu_coeff = Coefficient(mu, geometry, degree=degree)

        coefficients = epsilon_coeff, mu_coeff
        formulation = Maxwell2DBands(
            geometry,
            coefficients,
            function_space,
            propagation_vector=propagation_vector,
            degree=degree,
            polarization=polarization,
            boundary_conditions=boundary_conditions,
        )

        super().__init__(geometry, formulation)

        self.degree = degree
        self.propagation_vector = propagation_vector

    def eigensolve(self, *args, **kwargs):
        sol = super().eigensolve(*args, **kwargs)
        self.solution["eigenvectors"] = [
            u * self.formulation.phasor for u in sol["eigenvectors"]
        ]
        return self.solution



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io

# __all__ = ["colors", "pause", "plot", "plot_subdomains"]

import copy

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from dolfin.common.plotting import mesh2triang
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.transforms import Affine2D
from matplotlib.tri import Triangulation

from . import dolfin
from .complex import *
from .utils.helpers import project_iterative

colors = dict(
    red=(210 / 255, 95 / 255, 95 / 255), green=(69 / 255, 149 / 255, 125 / 255)
)

plt.register_cmap(
    cmap=LinearSegmentedColormap.from_list(
        "gyptis", [colors["green"], (1, 1, 1), colors["red"]], N=100
    )
)
plt.register_cmap(
    cmap=LinearSegmentedColormap.from_list(
        "gyptis_r", [colors["red"], (1, 1, 1), colors["green"]], N=100
    )
)
plt.register_cmap(
    cmap=LinearSegmentedColormap.from_list(
        "gyptis_white", [(1, 1, 1), (1, 1, 1), (1, 1, 1)], N=100
    )
)

plt.register_cmap(
    cmap=LinearSegmentedColormap.from_list(
        "gyptis_black", [(0, 0, 0), (0, 0, 0), (0, 0, 0)], N=100
    )
)


def get_boundaries(markers, domain=None, shift=(0, 0)):

    data = markers.array()
    triang = mesh2triang(markers.mesh())
    if domain == None:
        ids = np.unique(data)
    else:
        ids = [domain]

    triangulations = []
    for id in ids:
        triang_ = copy.deepcopy(triang)
        triang_.set_mask(data != id)
        triangulations.append(triang_)

    sub_bnds = []
    for triangtest in triangulations:
        maskedTris = triangtest.get_masked_triangles()
        verts = np.stack((triangtest.x[maskedTris], triangtest.y[maskedTris]), axis=-1)
        all_vert = np.vstack(verts).T
        sub_triang = Triangulation(*all_vert)

        boundaries = []
        for i in range(len(sub_triang.triangles)):
            for j in range(3):
                if sub_triang.neighbors[i, j] < 0:
                    boundaries.append(
                        (
                            sub_triang.triangles[i, j],
                            sub_triang.triangles[i, (j + 1) % 3],
                        )
                    )
        boundaries = np.asarray(boundaries)

        bndpnts = (
            shift[0] + sub_triang.x[boundaries].T,
            shift[1] + sub_triang.y[boundaries].T,
        )

        sub_bnds.append(bndpnts)

    return sub_bnds


def plot_boundaries(markers, domain=None, shift=(0, 0), ax=None, **kwargs):
    sub_bnds = get_boundaries(markers, domain=domain, shift=shift)
    if "c" not in kwargs and "color" not in kwargs:
        kwargs["color"] = "k"
    if ax is None:
        ax = plt.gca()
    l = []
    for bndpnts in sub_bnds:
        l_ = ax.plot(*bndpnts, **kwargs)
        l.append(l_)
    return l


# def plot_subdomains(markers, alpha=0.3):
#     a = dolfin.plot(markers, cmap="binary", alpha=alpha, lw=0.00, edgecolor="face")
#     return a
#     # a.set_edgecolors((0.1, 0.2, 0.5, 0.))


def plot_subdomains(markers, **kwargs):
    return plot_boundaries(markers, **kwargs)


def plot_real(fplot, ax=None, geometry=None, proj_space=None, colorbar=True, **kwargs):
    proj = proj_space is not None
    if "cmap" not in kwargs:
        kwargs["cmap"] = "RdBu_r"
    if ax is None:
        fig, ax = plt.subplots(1, 1)
    if proj:
        fplot = project_iterative(fplot, proj_space)
    plt.sca(ax)
    p = dolfin.plot(fplot, **kwargs)
    cbar = plt.colorbar(p) if colorbar else None
    kwargs.pop("cmap")
    if geometry:
        geometry.plot_subdomains(**kwargs)
    return p, cbar


def plot_complex(
    fplot, ax=None, geometry=None, proj_space=None, colorbar=True, **kwargs
):
    proj = proj_space is not None
    if ax is None:
        fig, ax = plt.subplots(1, 2)
    P, C = [], []
    for a, f in zip(ax, [fplot.real, fplot.imag]):
        p, cbar = plot_real(
            f,
            ax=a,
            geometry=geometry,
            proj_space=proj_space,
            colorbar=colorbar,
            **kwargs,
        )
        P.append(p)
        C.append(cbar)
    return P, C


def plot(fplot, **kwargs):
    if iscomplex(fplot):
        return plot_complex(fplot, **kwargs)
    else:
        return plot_real(fplot, **kwargs)


def pause(interval):
    backend = plt.rcParams["backend"]
    if backend in matplotlib.rcsetup.interactive_bk:
        figManager = matplotlib._pylab_helpers.Gcf.get_active()
        if figManager is not None:
            canvas = figManager.canvas
            if canvas.figure.stale:
                canvas.draw()
            canvas.start_event_loop(interval)
            return


def check_plot_type(plottype, f):
    if plottype == "real":
        fplot = f.real
    elif plottype == "imag":
        fplot = f.imag
    elif plottype == "module":
        fplot = f.module
    elif plottype == "phase":
        fplot = f.phase
    else:
        raise (
            ValueError(
                f"wrong plot type {plottype}, choose between real, imag, module or phase"
            )
        )
    return fplot



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io


from .source import *


def plane_wave_2d(wavelength, theta, amplitude=1, degree=1, domain=None):
    k0 = 2 * np.pi / wavelength
    K = k0 * np.array((-np.sin(theta), -np.cos(theta)))
    K_ = vector(sp.symbols("kx, ky, 0", real=True))
    expr = amplitude * sp.exp(1j * K_.dot(X))
    return expression2complex_2d(expr, kx=K[0], ky=K[1], degree=degree, domain=domain)


def plane_wave_3d(wavelength, theta, phi, psi, amplitude=1, degree=1, domain=None):

    cx = np.cos(psi) * np.cos(theta) * np.cos(phi) - np.sin(psi) * np.sin(phi)
    cy = np.cos(psi) * np.cos(theta) * np.sin(phi) + np.sin(psi) * np.cos(phi)
    cz = -np.cos(psi) * np.sin(theta)

    k0 = 2 * np.pi / wavelength
    K = k0 * np.array(
        (
            -np.sin(theta) * np.cos(phi),
            -np.sin(theta) * np.sin(phi),
            -np.cos(theta),
        )
    )
    K_ = vector(sp.symbols("kx, ky, kz", real=True))

    Propp = amplitude * sp.exp(1j * K_.dot(X))

    code = [sp.printing.ccode(p) for p in Propp.as_real_imag()]
    prop = dolfin.Expression(
        code, kx=K[0], ky=K[1], kz=K[2], degree=degree, domain=domain
    )
    C = dolfin.as_tensor([cx, cy, cz])
    return Complex(prop[0] * C, prop[1] * C)


class PlaneWave(Source):
    def __init__(self, wavelength, angle, dim, amplitude=1, degree=1, domain=None):
        super().__init__(wavelength, dim, degree=degree, domain=domain)
        self.angle = angle
        self.amplitude = amplitude

    @property
    def expression(self):
        if self.dim == 2:
            _expression = plane_wave_2d(
                self.wavelength,
                self.angle,
                amplitude=self.amplitude,
                degree=self.degree,
                domain=self.domain,
            )
        else:
            _expression = plane_wave_3d(
                self.wavelength,
                *self.angle,
                amplitude=self.amplitude,
                degree=self.degree,
                domain=self.domain,
            )
        return _expression



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io

from .source import *


class Dipole(Source):
    def __init__(
        self, wavelength, position, angle=0, amplitude=1, degree=1, domain=None
    ):
        super().__init__(wavelength, dim=2, degree=degree, domain=domain)
        self.position = position
        self.amplitude = amplitude
        self.angle = angle

    @property
    def expression(self):
        ls = LineSource(
            self.wavelength, self.position, self.amplitude, self.degree + 1, self.domain
        )
        n = as_vector([ConstantRe(-np.sin(self.angle)), ConstantRe(np.cos(self.angle))])
        dls = grad(ls.expression)
        _expression = dot(dls, n) / Constant(1j / self.wavenumber)
        return _expression



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io


from ..plot import *
from .metaclasses import _GratingBase
from .simulation import *


class Grating2D(_GratingBase, Simulation):
    def __init__(
        self,
        geometry,
        epsilon,
        mu,
        source=None,
        boundary_conditions={},
        polarization="TM",
        modal=False,
        degree=1,
        pml_stretch=1 - 1j,
        periodic_map_tol=1e-8,
        propagation_constant=0.0,
    ):
        assert isinstance(geometry, Layered2D)
        if source:
            assert isinstance(source, PlaneWave)
            assert source.dim == 2
        self.epsilon = epsilon
        self.mu = mu
        self.degree = degree
        self.period = geometry.period
        self.periodic_bcs = PeriodicBoundary2DX(self.period)
        function_space = ComplexFunctionSpace(
            geometry.mesh, "CG", degree, constrained_domain=self.periodic_bcs
        )
        pml_top = PML(
            "y",
            stretch=pml_stretch,
            matched_domain="superstrate",
            applied_domain="pml_top",
        )
        pml_bottom = PML(
            "y",
            stretch=pml_stretch,
            matched_domain="substrate",
            applied_domain="pml_bottom",
        )
        epsilon_coeff = Coefficient(
            epsilon,
            geometry=geometry,
            pmls=[pml_top, pml_bottom],
            degree=degree,
        )
        mu_coeff = Coefficient(
            mu, geometry=geometry, pmls=[pml_top, pml_bottom], degree=degree
        )
        coefficients = epsilon_coeff, mu_coeff
        no_source_domains = ["substrate", "superstrate", "pml_bottom", "pml_top"]

        if modal:
            source_domains = []
        else:
            source_domains = [
                dom for dom in geometry.domains if dom not in no_source_domains
            ]

        formulation = Maxwell2DPeriodic(
            geometry,
            coefficients,
            function_space,
            source=source,
            source_domains=source_domains,
            reference="superstrate",
            modal=modal,
            polarization=polarization,
            boundary_conditions=boundary_conditions,
            propagation_constant=propagation_constant,
            degree=degree,
        )
        super().__init__(geometry, formulation)

    def solve_system(self, again=False):
        uper = super().solve_system(again=again, vector_function=False)
        u_annex = self.formulation.annex_field["as_subdomain"]["stack"]
        u = uper * self.formulation.phasor
        self.solution = {}
        self.solution["periodic"] = uper
        self.solution["diffracted"] = u
        self.solution["total"] = u + u_annex
        return u

    def diffraction_efficiencies(
        self,
        N_order=0,
        cplx_effs=False,
        orders=False,
        subdomain_absorption=False,
        verbose=False,
    ):

        if self.formulation.polarization == "TM":
            nu = 1 / self.mu["substrate"]
        else:
            nu = 1 / self.epsilon["substrate"]
        orders_num = np.linspace(
            -N_order,
            N_order,
            2 * N_order + 1,
        )

        k, beta = {}, {}
        for d in ["substrate", "superstrate"]:
            k[d] = self.source.wavenumber * np.sqrt(
                complex(self.epsilon[d] * self.mu[d])
            )
            beta[d] = np.conj(
                np.sqrt(k[d] ** 2 - self.formulation.propagation_vector[0] ** 2)
            )

        phi_stack = self.formulation.annex_field["phi"]
        r_annex = phi_stack[0][-1]
        t_annex = phi_stack[-1][0]
        eff_annex = dict(substrate=t_annex, superstrate=r_annex)
        r_n, t_n = [], []
        R_n, T_n = [], []
        self.beta_n = []
        for n in orders_num:
            delta = 1 if n == 0 else 0
            qn = n * 2 * np.pi / self.period
            alpha_n = self.formulation.propagation_vector[0] + qn
            Jn, beta_n, eff = {}, {}, {}
            for d in ["substrate", "superstrate"]:
                s = 1 if d == "superstrate" else -1
                beta_n[d] = np.sqrt(k[d] ** 2 - alpha_n ** 2)
                ph_x = phasor(-qn, direction=0, degree=self.degree, domain=self.mesh)
                ph_y = phasor(
                    s * beta_n[d].real,
                    direction=1,
                    degree=self.degree,
                    domain=self.mesh,
                )
                Jn[d] = (
                    assemble(self.solution["periodic"] * ph_x * ph_y * self.dx(d))
                    / self.period
                )

                ph_pos = np.exp(-s * 1j * beta_n[d] * self.geometry.y_position[d])
                eff[d] = (
                    delta * eff_annex[d] + Jn[d] / self.geometry.thicknesses[d]
                ) * ph_pos
            r_, t_ = eff["superstrate"], eff["substrate"]
            sigma_r = beta_n["superstrate"] / beta["superstrate"]
            R_n_ = (r_ * r_.conj) * sigma_r
            sigma_t = beta_n["substrate"] / beta["superstrate"] * nu
            T_n_ = (t_ * t_.conj) * sigma_t

            self.beta_n.append(beta_n)

            r_n.append(r_)
            t_n.append(t_)
            R_n.append(R_n_.real)
            T_n.append(T_n_.real)

        R = np.sum(R_n, axis=0)
        T = np.sum(T_n, axis=0)

        Q, Qdomains = self.compute_absorption(subdomain_absorption=subdomain_absorption)

        B = T + R + Q  # energy balance

        if verbose:
            print(f"  Energy balance")
            print(f"  R = {R:0.6f}")
            print(f"  T = {T:0.6f}")
            print(f"  Q = {Q:0.6f}")
            print(f"  ------------------------")
            print(f"  B = {B:0.6f}")

        eff = dict()
        eff["R"] = r_n if cplx_effs else (R_n if orders else R)
        eff["T"] = t_n if cplx_effs else (T_n if orders else T)
        eff["Q"] = Qdomains if subdomain_absorption else Q
        eff["B"] = B
        self.efficiencies = eff
        return eff

    def compute_absorption(self, subdomain_absorption=False):
        omega = self.source.pulsation
        doms_no_pml = [
            z for z in self.epsilon.keys() if z not in ["pml_bottom", "pml_top"]
        ]

        xi_0, chi_0 = (
            (mu_0, epsilon_0)
            if self.formulation.polarization == "TM"
            else (epsilon_0, mu_0)
        )

        P0 = 0.5 * np.sqrt(chi_0 / xi_0) * np.cos(self.source.angle) * self.period

        Qchi = {}
        Qxi = {}
        u_tot = self.solution["total"]
        if subdomain_absorption:
            chi = self.formulation.chi.as_property()
            xi = self.formulation.xi.as_property()
            for d in doms_no_pml:

                nrj_chi_dens = (
                    dolfin.Constant(-0.5 * chi_0 * omega) * chi[d] * abs(u_tot) ** 2
                ).imag

                nrj_xi_dens = (
                    dolfin.Constant(-0.5 * 1 / (omega * xi_0))
                    * dot(grad(u_tot), (xi[d] * grad(u_tot)).conj).imag
                )

                Qchi[d] = assemble(nrj_chi_dens * self.dx(d)) / P0
                Qxi[d] = assemble(nrj_xi_dens * self.dx(d)) / P0
            Q = sum(Qxi.values()) + sum(Qchi.values())
        else:
            chi = self.formulation.chi.as_subdomain()
            xi = self.formulation.xi.as_subdomain()
            nrj_chi_dens = (
                dolfin.Constant(-0.5 * chi_0 * omega) * chi * abs(u_tot) ** 2
            ).imag

            nrj_xi_dens = (
                dolfin.Constant(-0.5 * 1 / (omega * xi_0))
                * dot(grad(u_tot), (xi * grad(u_tot)).conj).imag
            )
            Qchi = assemble(nrj_chi_dens * self.dx(doms_no_pml)) / P0
            Qxi = assemble(nrj_xi_dens * self.dx(doms_no_pml)) / P0
            Q = Qxi + Qchi

        if self.formulation.polarization == "TM":
            Qdomains = {"electric": Qchi, "magnetic": Qxi}
        else:
            Qdomains = {"electric": Qxi, "magnetic": Qchi}

        self.Qtot = Q
        self.Qdomains = Qdomains
        return Q, Qdomains

    def plot_geometry(self, nper=1, ax=None, c="k", **kwargs):

        if ax == None:
            ax = plt.gca()
        domains = self.geometry.subdomains["surfaces"]
        scatt = []
        for d in domains:
            if d not in self.geometry.layers:
                scatt.append(d)
        scatt_ids = [domains[d] for d in scatt]
        scatt_lines = []

        if len(scatt_ids) > 0:
            for i in range(nper):
                for sid in scatt_ids:
                    s = plot_subdomains(
                        self.geometry.markers,
                        domain=sid,
                        shift=(i * self.period, 0),
                        c=c,
                        **kwargs,
                    )
                    scatt_lines.append(s)
        yend = list(self.geometry.thicknesses.values())[-1]
        layers_lines = []
        for y0 in self.geometry.y_position.values():
            a = ax.axhline(y0, c=c, **kwargs)
            layers_lines.append(a)
        y0 += list(self.geometry.thicknesses.values())[-1]
        a = ax.axhline(y0, **kwargs)
        layers_lines.append(a)
        ax.set_aspect(1)
        return scatt_lines, layers_lines

    def plot_field(
        self,
        nper=1,
        type="real",
        field="total",
        fig=None,
        ax=None,
        mincmap=None,
        maxcmap=None,
        phase=0,
        callback=None,
        **kwargs,
    ):

        u = self.solution[field]
        if ax == None:
            ax = plt.gca()
        if "cmap" not in kwargs:
            kwargs["cmap"] = "RdBu_r"
        per_plots = []
        ppmin, ppmax = [], []
        for i in range(nper):
            alpha = self.formulation.propagation_vector[0]
            t = Affine2D().translate(i * self.period, 0)
            f = u * phase_shift(i * alpha * self.period + phase, degree=self.degree)
            fplot = check_plot_type(type, f)
            # if ADJOINT:
            fplot = project(
                fplot,
                self.formulation.real_function_space,
                solver_type="cg",
                preconditioner_type="jacobi",
            )
            plt.sca(ax)
            pp = dolfin.plot(fplot, transform=t + ax.transData, **kwargs)
            per_plots.append(pp)

            ppmax.append(pp.cvalues[-1])
            ppmin.append(pp.cvalues[0])
        ax.set_xlim(-self.period / 2, (nper - 1 / 2) * self.period)
        ax.set_aspect(1)
        mincmap = mincmap or min(ppmin)
        maxcmap = maxcmap or max(ppmax)
        for pp in per_plots:
            pp.set_clim(mincmap, maxcmap)

        cm = plt.cm.ScalarMappable(cmap=kwargs["cmap"])
        cm.set_clim(mincmap, maxcmap)

        fig = plt.gcf() if fig is None else fig
        cb = fig.colorbar(cm, ax=ax)

        if callback is not None:
            callback(**kwargs)

        return per_plots, cb

    def animate_field(self, n=11, filename="animation.gif", **kwargs):
        import tempfile

        from PIL import Image

        from ..plot import plt

        anim = []
        tmpdir = tempfile.mkdtemp()
        fp_in = f"{tmpdir}/animation_tmp_*.png"
        phase = np.linspace(0, 2 * np.pi, n + 1)[:n]
        for iplot in range(n):
            number_str = str(iplot).zfill(4)
            pngname = f"{tmpdir}/animation_tmp_{number_str}.png"
            p = self.plot_field(phase=phase[iplot], **kwargs)
            fig = plt.gcf()
            fig.savefig(pngname)
            fig.clear()
            anim.append(p)

        plt.close(fig)

        img, *imgs = [Image.open(f) for f in sorted(glob.glob(fp_in))]
        img.save(
            fp=filename,
            format="GIF",
            append_images=imgs,
            save_all=True,
            duration=200,
            loop=0,
        )
        os.system(f"rm -f {tmpdir}/animation_tmp_*.png")
        return anim

    def eigensolve(self, *args, **kwargs):
        sol = super().eigensolve(*args, **kwargs)
        self.solution["eigenvectors"] = [
            u * self.formulation.phasor for u in sol["eigenvectors"]
        ]
        return self.solution



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io


import os
from math import e, pi

import dolfin
from scipy.constants import c, epsilon_0, mu_0

from .__about__ import __author__, __description__, __version__

# __doc__ = __description__


if os.environ.get("GYPTIS_ADJOINT") is not None:
    import dolfin_adjoint

    ADJOINT = True
    dolfin.__dict__.update(dolfin_adjoint.__dict__)
    dolfin.__spec__.name = "dolfin"
    del dolfin_adjoint
    ## prevents AttributeError: module 'fenics_adjoint.types.function'
    ## has no attribute 'function' when writing solution to files with <<
    dolfin.function.function = dolfin.function
else:
    ADJOINT = False


from dolfin import MPI

dolfin.parameters["reorder_dofs_serial"] = False
dolfin.parameters["ghost_mode"] = "shared_facet"
dolfin.PETScOptions.set("petsc_prealloc", "200")
dolfin.PETScOptions.set("ksp_type", "preonly")
dolfin.PETScOptions.set("pc_type", "lu")
# dolfin.PETScOptions.set('pc_factor_mat_solver_type', 'mumps')
# dolfin.PETScOptions.set('sub_pc_type', 'lu')
# dolfin.PETScOptions.set('pc_asm_overlap', '10')
# dolfin.PETScOptions.set('ksp_type', 'gmres')
# dolfin.PETScOptions.set('petsc_prealloc', '100')
# dolfin.PETScOptions.set('ksp_rtol', '1.e-12')
# dolfin.PETScOptions.set('ksp_type', 'gmres')
# dolfin.PETScOptions.set('ksp_gmres_restart', '100')
# dolfin.PETScOptions.set('pc_type', 'jacobi')
# dolfin.PETScOptions.set('pc_factor_levels', '3')
# dolfin.PETScOptions.set('mat_mumps_icntl_14', 40)
# dolfin.PETScOptions.set('mat_mumps_icntl_35', 2)
# dolfin.PETScOptions.set('mat_mumps_icntl_36', 1)
# dolfin.PETScOptions.set('mat_mumps_icntl_24', 1)
# dolfin.set_log_level(10000000)

from .api import *
from .complex import *
from .plot import *
from .utils import logger, set_log_level

# logger.warning("Welcome to gyptis!")
# logger.info("Welcome to gyptis!")
# logger.debug("Welcome to gyptis!")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io

from .geometry import *


class BoxPML3D(Geometry):
    def __init__(
        self,
        box_size=(1, 1, 1),
        box_center=(0, 0, 0),
        pml_width=(0.2, 0.2, 0.2),
        Rcalc=0,
        **kwargs,
    ):
        super().__init__(
            dim=3,
            **kwargs,
        )
        self.box_size = box_size
        self.box_center = box_center
        self.pml_width = pml_width
        self.Rcalc = Rcalc

        box = self._addbox_center(self.box_size)
        T = np.array(self.pml_width) / 2 + np.array(self.box_size) / 2

        s = (self.pml_width[0], self.box_size[1], self.box_size[2])
        t = np.array([T[0], 0, 0])
        pmlxp = self._add_pml(s, t)
        pmlxm = self._add_pml(s, -t)

        s = (self.box_size[0], self.pml_width[1], self.box_size[2])
        t = np.array([0, T[1], 0])
        pmlyp = self._add_pml(s, t)
        pmlym = self._add_pml(s, -t)

        s = (self.box_size[0], self.box_size[1], self.pml_width[2])
        t = np.array([0, 0, T[2]])
        pmlzp = self._add_pml(s, t)
        pmlzm = self._add_pml(s, -t)

        s = (self.pml_width[0], self.pml_width[1], self.box_size[2])
        pmlxypp = self._add_pml(s, [T[0], T[1], 0])
        pmlxypm = self._add_pml(s, [T[0], -T[1], 0])
        pmlxymp = self._add_pml(s, [-T[0], T[1], 0])
        pmlxymm = self._add_pml(s, [-T[0], -T[1], 0])

        s = (self.box_size[0], self.pml_width[1], self.pml_width[2])
        pmlyzpp = self._add_pml(s, [0, T[1], T[2]])
        pmlyzpm = self._add_pml(s, [0, T[1], -T[2]])
        pmlyzmp = self._add_pml(s, [0, -T[1], T[2]])
        pmlyzmm = self._add_pml(s, [0, -T[1], -T[2]])

        s = (self.pml_width[0], self.box_size[1], self.pml_width[2])
        pmlxzpp = self._add_pml(s, [T[0], 0, T[2]])
        pmlxzpm = self._add_pml(s, [T[0], 0, -T[2]])
        pmlxzmp = self._add_pml(s, [-T[0], 0, T[2]])
        pmlxzmm = self._add_pml(s, [-T[0], 0, -T[2]])

        s = (self.pml_width[0], self.pml_width[1], self.pml_width[2])
        pmlxyzppp = self._add_pml(s, [T[0], T[1], T[2]])
        pmlxyzppm = self._add_pml(s, [T[0], T[1], -T[2]])
        pmlxyzpmp = self._add_pml(s, [T[0], -T[1], T[2]])
        pmlxyzpmm = self._add_pml(s, [T[0], -T[1], -T[2]])
        pmlxyzmpp = self._add_pml(s, [-T[0], T[1], T[2]])
        pmlxyzmpm = self._add_pml(s, [-T[0], T[1], -T[2]])
        pmlxyzmmp = self._add_pml(s, [-T[0], -T[1], T[2]])
        pmlxyzmmm = self._add_pml(s, [-T[0], -T[1], -T[2]])

        pmlx = [pmlxp, pmlxm]
        pmly = [pmlyp, pmlym]
        pmlz = [pmlzp, pmlzm]
        pml1 = pmlx + pmly + pmlz

        pmlxy = [pmlxypp, pmlxypm, pmlxymp, pmlxymm]
        pmlyz = [pmlyzpp, pmlyzpm, pmlyzmp, pmlyzmm]
        pmlxz = [pmlxzpp, pmlxzpm, pmlxzmp, pmlxzmm]
        pml2 = pmlxy + pmlyz + pmlxz

        pml3 = [
            pmlxyzppp,
            pmlxyzppm,
            pmlxyzpmp,
            pmlxyzpmm,
            pmlxyzmpp,
            pmlxyzmpm,
            pmlxyzmmp,
            pmlxyzmmm,
        ]

        self.box = box
        self.pmls = pml1 + pml2 + pml3
        self._translate([self.box] + self.pmls, self.box_center)
        self.fragment(self.box, self.pmls)

        if Rcalc > 0:
            sphere_calc = self.add_sphere(*self.box_center, Rcalc)
            box, sphere_calc = self.fragment(box, sphere_calc)
            self.box = [box, sphere_calc]

        self.add_physical(box, "box")
        self.add_physical(pmlx, "pmlx")
        self.add_physical(pmly, "pmly")
        self.add_physical(pmlz, "pmlz")
        self.add_physical(pmlxy, "pmlxy")
        self.add_physical(pmlyz, "pmlyz")
        self.add_physical(pmlxz, "pmlxz")
        self.add_physical(pml3, "pmlxyz")

        self.pml_physical = [
            "pmlx",
            "pmly",
            "pmlz",
            "pmlxy",
            "pmlyz",
            "pmlxz",
            "pmlxyz",
        ]

        if Rcalc > 0:
            bnds = self.get_boundaries("box")
            self.calc_bnds = bnds[0]
            self.add_physical(self.calc_bnds, "calc_bnds", dim=2)

    def _addbox_center(self, rect_size):
        corner = -np.array(rect_size) / 2
        corner = tuple(corner)
        return self.add_box(*corner, *rect_size)

    def _translate(self, tag, t):
        translation = tuple(t)
        self.translate(self.dimtag(tag), *translation)

    def _add_pml(self, s, t):
        pml = self._addbox_center(s)
        self._translate(pml, t)
        return pml



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io

"""
Support for complex finite element forms.
This module provides a :class:`Complex` class and overrides some ``dolfin`` functions
to easily deal with complex problems by spliting real and imaginary parts.
"""

from typing import Iterable

import numpy as np
import ufl

from . import dolfin


def _complexcheck(func):
    """Wrapper to check if arguments are complex"""

    def wrapper(self, z):
        if hasattr(z, "real") and hasattr(z, "imag"):
            if not isinstance(z, Complex):
                z = Complex(z.real, z.imag)
        else:
            z = Complex(z, 0)
        return func(self, z)

    return wrapper


def _complexify_linear(func):
    def wrapper(z, *args, **kwargs):
        if isinstance(z, Complex):
            return Complex(func(z.real, *args, **kwargs), func(z.imag, *args, **kwargs))
        else:
            return func(z, *args, **kwargs)

    return wrapper


def _complexify_bilinear(func):
    def wrapper(a, b, *args, **kwargs):
        if iscomplex(a) and iscomplex(b):
            re = func(a.real, b.real, *args, **kwargs) - func(
                a.imag, b.imag, *args, **kwargs
            )
            im = func(a.real, b.imag, *args, **kwargs) + func(
                a.imag, b.real, *args, **kwargs
            )
            return Complex(re, im)
        elif iscomplex(a) and not iscomplex(b):
            re = func(a.real, b, *args, **kwargs)
            im = func(a.imag, b)
            return Complex(re, im, *args, **kwargs)
        elif not iscomplex(a) and iscomplex(b):
            re = func(a, b.real, *args, **kwargs)
            im = func(a, b.imag)
            return Complex(re, im, *args, **kwargs)
        else:
            return func(a, b, *args, **kwargs)

    return wrapper


def _complexify_vector(func):
    def wrapper(*args, **kwargs):
        v = func(*args, **kwargs)
        re, im = dolfin.split(v)
        return Complex(re, im)

    return wrapper


def _complexify_vector_alt(func):
    def wrapper(*args, **kwargs):
        v = func(*args, **kwargs)
        re, im = v.split()  # (deepcopy=True)
        return Complex(re, im)

    return wrapper


class Complex:
    """A complex object.

    Parameters
    ----------
    real : type
        Real part.
    imag : type
        Imaginary part (the default is 0.0).

    Attributes
    ----------
    real
    imag

    """

    def __init__(self, real, imag=0.0):
        self.real = real
        self.imag = imag

    def __len__(self):
        if hasattr(self.real, "__len__") and hasattr(self.imag, "__len__"):
            if len(self.real) == len(self.imag):
                return len(self.real)
            else:
                raise ValueError("real and imaginary parts should have the same length")
        else:
            return 0

    def __iter__(self):
        for i in range(len(self)):
            yield Complex(self.real[i], self.imag[i])

    def __getitem__(self, i):
        return Complex(self.real[i], self.imag[i])

    @_complexcheck
    def __add__(self, other):
        return Complex(self.real + other.real, self.imag + other.imag)

    __radd__ = __add__

    @_complexcheck
    def __sub__(self, other):
        return Complex(self.real - other.real, self.imag - other.imag)

    __rsub__ = __sub__

    @_complexcheck
    def __mul__(self, other):
        return Complex(
            self.real * other.real - self.imag * other.imag,
            self.imag * other.real + self.real * other.imag,
        )

    __rmul__ = __mul__

    __array_ufunc__ = None

    @_complexcheck
    def __truediv__(self, other):
        sr, si, tr, ti = self.real, self.imag, other.real, other.imag  # short forms
        r = tr ** 2 + ti ** 2
        return Complex((sr * tr + si * ti) / r, (si * tr - sr * ti) / r)

    @_complexcheck
    def __rtruediv__(self, other):
        sr, si, tr, ti = other.real, other.imag, self.real, self.imag  # short forms
        r = tr ** 2 + ti ** 2
        return Complex((sr * tr + si * ti) / r, (si * tr - sr * ti) / r)

    @property
    def shape(self):
        return self.real.shape

    @property
    def conj(self):
        return Complex(self.real, -self.imag)

    @property
    def module(self):
        """Module of the complex number"""
        return self.__abs__()

    @property
    def phase(self):
        return self.__angle__()

    def __abs__(self):
        return dolfin.sqrt(self.real ** 2 + self.imag ** 2)

    def __neg__(self):  # defines -c (c is Complex)
        return Complex(-self.real, -self.imag)

    @_complexcheck
    def __eq__(self, other):
        return self.real == other.real and self.imag == other.imag

    @_complexcheck
    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return f"({self.real.__str__()} + {self.imag.__str__()}j)"

    def __repr__(self):
        return "Complex" + f"({self.real.__repr__()}, {self.imag.__repr__()})"

    def __pow__(self, power):
        if iscomplex(power) and power.imag != 0:
            raise NotImplementedError("complex exponent not implemented")
        else:
            A, phi = self.polar()
            return self.polar2cart(A ** power, phi * power)

    def __angle__(self):
        x, y = self.real, self.imag
        try:
            return np.angle(x + 1j * y)
        except:
            return dolfin.conditional(
                ufl.eq(self.__abs__(), 0),
                0,
                dolfin.conditional(
                    ufl.eq(self.__abs__() + x, 0),
                    dolfin.pi,
                    2 * dolfin.atan(y / (self.__abs__() + x)),
                ),
            )

    def __call__(self, *args, **kwargs):
        "Calls the complex function if base objects are callable"
        return Complex(
            self.real.__call__(*args, **kwargs),
            self.imag.__call__(*args, **kwargs),
        )

    def tocomplex(self):
        return self.real + 1j * self.imag

    @staticmethod
    def polar2cart(module, phase):
        """Polar to cartesian representation.

        Parameters
        ----------
        module : type
            The module (positive).
        phase : type
            The polar angle.

        Returns
        -------
        Complex
            The complex number in cartesian representation.

        """
        return module * Complex(dolfin.cos(phase), dolfin.sin(phase))

    def polar(self):
        """Polar representation.

        Returns
        -------
        tuple
            Modulus and phase.

        """
        return self.__abs__(), self.__angle__()

    def to_complex(self):
        return self.real + 1j * self.imag


def iscomplex(z):
    """Checks if object is complex.

    Parameters
    ----------
    z : type
        Object.

    Returns
    -------
    bool
        True if z is complex, else False.

    """
    if hasattr(z, "real") and hasattr(z, "imag") and z.imag is not None:
        return True
    else:
        return False


class ComplexFunctionSpace(dolfin.FunctionSpace):
    """Complex function space"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        element = super().ufl_element()
        super().__init__(super().mesh(), element * element, **kwargs)


def _cplx_iter(f):
    def wrapper(v, *args, **kwargs):
        iterable = isinstance(v, Iterable)
        cplx = any([iscomplex(v_) for v_ in v]) if iterable else iscomplex(v)
        if cplx:
            if iterable:
                v = np.array(v)
            return Complex(f(v.real, *args, **kwargs), f(v.imag, *args, **kwargs))
        else:
            return f(v, *args, **kwargs)

    return wrapper


def phasor(prop_cst, direction=0, **kwargs):
    phasor_re = dolfin.Expression(
        f"cos(prop_cst*x[{direction}])", prop_cst=prop_cst, **kwargs
    )
    phasor_im = dolfin.Expression(
        f"sin(prop_cst*x[{direction}])", prop_cst=prop_cst, **kwargs
    )
    return Complex(phasor_re, phasor_im)


def phase_shift(phase, **kargs):
    phasor_re = dolfin.Expression(f"cos(phase)", phase=phase, **kargs)
    phasor_im = dolfin.Expression(f"sin(phase)", phase=phase, **kargs)
    return Complex(phasor_re, phasor_im)


def vector(vect):
    vsr = [v.real for v in vect]
    vsi = [v.imag for v in vect]
    return Complex(dolfin.as_vector(vsr), dolfin.as_vector(vsi))


def tensor(tens):
    tsr = [[t.real for t in _t] for _t in tens]
    tsi = [[t.imag for t in _t] for _t in tens]
    return Complex(dolfin.as_tensor(tsr), dolfin.as_tensor(tsi))


interpolate = _complexify_linear(dolfin.interpolate)
assemble = _complexify_linear(dolfin.assemble)
grad = _complexify_linear(dolfin.grad)
div = _complexify_linear(dolfin.div)
curl = _complexify_linear(dolfin.curl)
project = _complexify_linear(dolfin.project)

Dx = _complexify_linear(dolfin.Dx)
inner = _complexify_bilinear(dolfin.inner)
dot = _complexify_bilinear(dolfin.dot)
cross = _complexify_bilinear(dolfin.cross)
as_tensor = _cplx_iter(dolfin.as_tensor)
as_vector = _cplx_iter(dolfin.as_vector)
Constant = _cplx_iter(dolfin.Constant)
Function = _complexify_vector_alt(dolfin.Function)
TrialFunction = _complexify_vector(dolfin.TrialFunction)
TestFunction = _complexify_vector(dolfin.TestFunction)
TrialFunctions = _complexify_vector(dolfin.TrialFunctions)
TestFunctions = _complexify_vector(dolfin.TestFunctions)
j = Complex(0, 1)


def _traverse(a):
    if not isinstance(a, list):
        yield a
    else:
        for e in a:
            yield from _traverse(e)


def to_array(a):
    l = []
    for e in _traverse(a):
        l.append(e.to_complex())
    return np.array(l).reshape(np.array(a).shape[:-1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io


from .formulation import *


class Maxwell3D(Formulation):
    def __init__(
        self,
        geometry,
        coefficients,
        function_space,
        source=None,
        boundary_conditions={},
        source_domains=[],
        reference=None,
        modal=False,
        degree=1,
    ):
        super().__init__(
            geometry,
            coefficients,
            function_space,
            source=source,
            boundary_conditions=boundary_conditions,
            modal=modal,
            degree=degree,
            dim=3,
        )

        self.source_domains = source_domains
        self.reference = reference
        self.epsilon, self.mu = self.coefficients
        self.pec_boundaries = prepare_boundary_conditions(boundary_conditions)

    def maxwell(self, u, v, epsilon, inv_mu, domain="everywhere"):

        form = []
        form.append(-inner(inv_mu * curl(u), curl(v)))
        form.append(inner(epsilon * u, v))
        if self.modal:
            return [form[0] * self.dx(domain), -form[1] * self.dx(domain)]
        else:
            k0 = Constant(self.source.wavenumber)
            return (form[0] + k0 ** 2 * form[1]) * self.dx(domain)

    def _weak(self, u, v, u1):
        epsilon = self.epsilon.as_subdomain()
        inv_mu = self.mu.invert().as_subdomain()

        epsilon_dict = self.epsilon.as_property()
        inv_mu_dict = self.mu.invert().as_property()

        dom_func, dom_no_func = find_domains_function((self.epsilon, self.mu))
        source_dom_func, source_dom_no_func = find_domains_function(
            (self.epsilon, self.mu), self.source_domains
        )
        form = self.maxwell(u, v, epsilon, inv_mu, domain=dom_no_func)

        for dom in dom_func:
            if self.modal:
                form_dom_func = self.maxwell(
                    u, v, xi_dict[dom], chi_dict[dom], domain=dom
                )
                form = [form[i] + form_dom_func[i] for i in range(2)]
            else:
                form += self.maxwell(
                    u, v, epsilon_dict[dom], inv_mu_dict[dom], domain=dom
                )

        if self.source_domains != []:
            epsilon_a = self.epsilon.build_annex(
                domains=self.source_domains, reference=self.reference
            ).as_subdomain()
            inv_mu_a = (
                self.mu.invert()
                .build_annex(domains=self.source_domains, reference=self.reference)
                .as_subdomain()
            )
            epsilon_a_dict = self.epsilon.build_annex(
                domains=self.source_domains, reference=self.reference
            ).as_property()
            inv_mu_a_dict = (
                self.mu.invert()
                .build_annex(domains=self.source_domains, reference=self.reference)
                .as_property()
            )
            if source_dom_no_func != []:
                form += self.maxwell(
                    u1,
                    v,
                    epsilon - epsilon_a,
                    inv_mu - inv_mu_a,
                    domain=source_dom_no_func,
                )
            for dom in source_dom_func:
                form += self.maxwell(
                    u1,
                    v,
                    epsilon_dict[dom] - epsilon_a_dict[dom],
                    inv_mu_dict[dom] - inv_mu_a_dict[dom],
                    domain=dom,
                )
        if self.modal:
            weak = [f.real + f.imag for f in form]
        else:
            weak = form.real + form.imag
        return weak

    @property
    def weak(self):
        u1 = self.source.expression
        u = self.trial
        v = self.test
        return self._weak(u, v, u1)

    def build_pec_boundary_conditions(self, applied_function):
        if self.pec_boundaries != []:
            ## FIXME: project is slow, avoid it.
            applied_function = project_iterative(
                applied_function, self.real_function_space
            )
            _boundary_conditions = build_pec_boundary_conditions(
                self.pec_boundaries,
                self.geometry,
                self.function_space,
                applied_function,
            )
        else:
            _boundary_conditions = []
        return _boundary_conditions

    def build_boundary_conditions(self):
        applied_function = (
            Constant((0, 0, 0)) if self.modal else -self.source.expression
        )
        self._boundary_conditions = self.build_pec_boundary_conditions(applied_function)
        return self._boundary_conditions



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io


from .maxwell3d import *


class Maxwell3DBands(Maxwell3D):
    def __init__(self, *args, propagation_vector=(0, 0, 0), **kwargs):
        super().__init__(*args, **kwargs, modal=True)
        self.propagation_vector = propagation_vector

    @property
    def phasor_vect(self):

        return [
            phasor(
                self.propagation_vector[i],
                direction=i,
                degree=self.degree,
                domain=self.geometry.mesh,
            )
            for i in range(3)
        ]

    @property
    def phasor(self):
        return self.phasor_vect[0] * self.phasor_vect[1] * self.phasor_vect[2]

    @property
    def weak(self):
        u = self.trial * self.phasor
        v = self.test * self.phasor.conj
        return super()._weak(u, v, Constant((0, 0, 0)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io

from .simulation import *


class Homogenization2D(Simulation):
    def __init__(
        self,
        geometry,
        epsilon,
        mu,
        boundary_conditions={},
        degree=1,
        direction="x",
        direct=True,
    ):
        assert isinstance(geometry, Lattice2D)

        self.periodic_bcs = BiPeriodic2D(geometry)
        function_space = ComplexFunctionSpace(
            geometry.mesh, "CG", degree, constrained_domain=self.periodic_bcs
        )
        epsilon = {k: e + 1e-16j for k, e in epsilon.items()}
        mu = {k: m + 1e-16j for k, m in mu.items()}
        epsilon_coeff = Coefficient(epsilon, geometry, degree=degree)
        mu_coeff = Coefficient(mu, geometry, degree=degree)

        coefficients = epsilon_coeff, mu_coeff
        self.formulations = dict(epsilon={}, mu={})
        for case in ["epsilon", "mu"]:
            for direction in ["x", "y"]:
                form = TwoScale2D(
                    geometry,
                    coefficients,
                    function_space,
                    degree=degree,
                    boundary_conditions=boundary_conditions,
                    direction=direction,
                    case=case,
                )
                self.formulations[case][direction] = form
        self.formulation = self.formulations["epsilon"]["x"]
        super().__init__(geometry, self.formulation, direct=direct)
        self.solution = dict(epsilon={}, mu={})
        self.direct = direct
        self.direction = direction
        self.degree = degree
        self.cell_volume = np.cross(*geometry.vectors)

    def solve_system(self, again=False):
        return super().solve_system(again=again, vector_function=False)

    def unit_cell_mean(self, f):
        return 1 / self.cell_volume * assemble(f * self.dx)

    def solve_param(self, case):
        self.formulation = self.formulations[case]["x"]
        super().__init__(self.geometry, self.formulation, direct=self.direct)
        phi_x = self.solve()
        self.formulations[case]["y"].build_rhs()
        self.vector = assemble(self.formulations[case]["y"].rhs)
        phi_y = self.solve_system(again=True)
        self.solution[case] = dict(x=phi_x, y=phi_y)
        return self.solution

    def solve_all(self):
        phi_x = self.solve()
        self.formulation_y.build_rhs()
        self.vector = assemble(self.formulation_y.rhs)
        phi_y = self.solve_system(again=True)
        self.solution = dict(x=phi_x, y=phi_y)
        return self.solution

    def _get_effective_coeff(self, case):
        self.solve_param(case)
        xi = self.formulation.xi.as_subdomain()
        if xi.real.ufl_shape == (2, 2):
            xi_mean = []
            for i in range(2):
                a = [self.unit_cell_mean(x) for x in xi[i]]
                a = [_.real + 1j * _.imag for _ in a]
                xi_mean.append(a)
            xi_mean = np.array(xi_mean)
        else:
            xi_mean = self.unit_cell_mean(xi)
            xi_mean = xi_mean.real + 1j * xi_mean.imag
            xi_mean *= np.eye(2)
        A = []
        for phi in self.solution[case].values():
            integrand = xi * grad(phi)
            a = [self.unit_cell_mean(g) for g in integrand]
            a = [_.real + 1j * _.imag for _ in a]
            A.append(a)
        xi_eff = xi_mean + np.array(A)
        param_eff_inplane = xi_eff.T / np.linalg.det(xi_eff)
        param_eff = np.zeros((3, 3), complex)
        param_eff[:2, :2] = param_eff_inplane
        i = 0 if case == "epsilon" else 1
        coeff = self.formulation.coefficients[i].as_subdomain()
        if coeff.real.ufl_shape == (3, 3):
            coeffzz = coeff[2][2]
        else:
            coeffzz = coeff
        czz = self.unit_cell_mean(coeffzz)
        param_eff[2, 2] = czz.real + 1j * czz.imag
        return param_eff

    def get_effective_permittivity(self):
        return self._get_effective_coeff("epsilon")

    def get_effective_permeability(self):
        return self._get_effective_coeff("mu")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io

from .geometry import *


class Layered3D(Geometry):
    def __init__(
        self,
        period=(1.0, 1.0),
        thicknesses=None,
        **kwargs,
    ):
        super().__init__(
            dim=3,
            **kwargs,
        )
        self.period = period
        self.thicknesses = thicknesses or OrderedDict(
            {
                "pml_bottom": 1,
                "substrate": 1,
                "groove": 2,
                "superstrate": 1,
                "pml_top": 1,
            }
        )

        self.periodic_tol = 1e-6

        self.translation_x = self._translation_matrix([self.period[0], 0, 0])
        self.translation_y = self._translation_matrix([0, self.period[1], 0])

        self.total_thickness = sum(self.thicknesses.values())
        dx, dy = self.period
        self.z0 = -sum(list(self.thicknesses.values())[:2])

        z0 = self.z0
        self.layers = {}
        self.z_position = {}
        for id, thickness in self.thicknesses.items():
            layer = self.make_layer(z0, thickness)
            self.layers[id] = layer
            self.z_position[id] = z0
            self.add_physical(layer, id)
            z0 += thickness

        self.remove_all_duplicates()
        self.synchronize()
        for sub, num in self.subdomains["volumes"].items():
            self.add_physical(num, sub)

    def make_layer(self, z_position, thickness):
        return self.add_box(
            -self.period[0] / 2,
            -self.period[1] / 2,
            z_position,
            self.period[0],
            self.period[1],
            thickness,
        )

    def set_periodic_mesh(self, eps=None):
        s = self.get_periodic_bnds(self.z0, self.total_thickness, eps=eps)

        periodic_id = {}
        for k, v in s.items():
            periodic_id[k] = [S[-1] for S in v]
        gmsh.model.mesh.setPeriodic(
            2, periodic_id["+x"], periodic_id["-x"], self.translation_x
        )
        gmsh.model.mesh.setPeriodic(
            2, periodic_id["+y"], periodic_id["-y"], self.translation_y
        )
        self.periodic_bnds = periodic_id
        return periodic_id

    def build(self, set_periodic=True, **kwargs):
        if set_periodic:
            self.set_periodic_mesh()
        super().build(**kwargs)

    def get_periodic_bnds(self, z_position, thickness, eps=None):

        eps = eps or self.periodic_tol

        s = {}
        dx, dy = self.period

        pmin = -dx / 2 - eps, -dy / 2 - eps, z_position - eps
        pmax = -dx / 2 + eps, +dy / 2 + eps, thickness + eps
        s["-x"] = gmsh.model.getEntitiesInBoundingBox(*pmin, *pmax, 2)

        pmin = +dx / 2 - eps, -dy / 2 - eps, z_position - eps
        pmax = +dx / 2 + eps, +dy / 2 + eps, thickness + eps
        s["+x"] = gmsh.model.getEntitiesInBoundingBox(*pmin, *pmax, 2)

        pmin = -dx / 2 - eps, -dy / 2 - eps, z_position - eps
        pmax = +dx / 2 + eps, -dy / 2 + eps, thickness + eps
        s["-y"] = gmsh.model.getEntitiesInBoundingBox(*pmin, *pmax, 2)

        pmin = -dx / 2 - eps, +dy / 2 - eps, z_position - eps
        pmax = +dx / 2 + eps, +dy / 2 + eps, thickness + eps
        s["+y"] = gmsh.model.getEntitiesInBoundingBox(*pmin, *pmax, 2)
        return s



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of gyptis
# License: MIT
# See the documentation at gyptis.gitlab.io


from . import dolfin


class Measure(dolfin.Measure):
    def __init__(
        self,
        integral_type,
        domain=None,
        subdomain_id="everywhere",
        metadata=None,
        subdomain_data=None,
        subdomain_dict=None,
    ):

        self.subdomain_dict = subdomain_dict
        if (
            self.subdomain_dict
            and isinstance(subdomain_id, str)
            and subdomain_id != "everywhere"
        ):
            subdomain_id = self.subdomain_dict[subdomain_id]
        super().__init__(
            integral_type,
            domain=domain,
            subdomain_id=subdomain_id,
            metadata=metadata,
            subdomain_data=subdomain_data,
        )

    def __call_single__(self, subdomain_id=None, **kwargs):
        if (
            self.subdomain_dict
            and isinstance(subdomain_id, str)
            and subdomain_id != "everywhere"
        ):
            subdomain_id = self.subdomain_dict[subdomain_id]
        return super().__call__(subdomain_id=subdomain_id, **kwargs)

    def __call__(self, subdomain_id=None, **kwargs):
        subdomain_id = None if subdomain_id == [] else subdomain_id
        if isinstance(subdomain_id, list):
            for i, sid in enumerate(subdomain_id):
                if i == 0:
                    out = self.__call_single__(subdomain_id=sid, **kwargs)
                else:
                    out += self.__call_single__(subdomain_id=sid, **kwargs)
            return out
        else:
            return self.__call_single__(subdomain_id=subdomain_id, **kwargs)
